from itertools import product, chain
import operator
import romkan
from romkan import to_hiragana, to_kunrei
from .re_one_of import re_one_of
import re

vow = 'aiueo'
vow2 = tuple(vow) + ('ya', 'yo', 'yu')
con = 'kgsztdnhbpmr'
no_sokuon = set(vow2)
all_mora = set(map(lambda x: operator.add(*x),
                   product(con+'yw', vow)))
all_mora.update(vow)
all_mora.add("n'")
all_mora.add('xtu')
all_mora.add('vu')
all_mora.difference_update(('yi', 'ye', 'wi', 'wu', 'we'))
all_mora.update(p+'x'+v for v in vow2
                for p in chain((c+'i' for c in con), ('vu',)))
all_mora = dict((k, k) for k in all_mora)

all_mora = {k:to_hiragana(v).replace('う゛','ゔ')
            for k, v in all_mora.items()}
all_mora.update({'wi':'ゐ', 'we':'ゑ'})

def _make_conjugation_table():
    d = {}
    v5c = 'trkgbms'

    def e(form, *args):
        for arg in args:
            d.update(((form, k), v) for k, v in arg)

    e('polite', (('v1', ''), ('v5u', 'i')),
      (('v5'+k, k+'i') for k in v5c))

    e('negative', (('v1', 'nai'), ('v5u', 'wanai')),
      (('v5'+k, k+'anai') for k in v5c))

    for v, n in (('e', 'te'), ('a', 'perfective')):
        e(n,
          (('v1', 't'+v), ('v5k', 'it'+v), ('v5g', 'id'+v), ('v5s', 'sit'+v)),
          (('v5'+k, 'xtut'+v) for k in 'utr'),
          (('v5'+k,  "n'd"+v) for k in 'bm'))

    e('potential', (('v1', 'rareru'), ('v5u', 'eru')),
      (('v5'+k, k+'eru') for k in v5c))

    e('conditional', (('v1', 'reba'), ('v5u', 'eba')),
      (('v5'+k, k+'eba') for k in v5c))

    e('volitional', (('v1', 'you'), ('v5u', 'ou')),
      (('v5'+k, k+'ou') for k in v5c))

    e('passive',   (('v1', 'rareru'), ('v5u', 'wareru')),
      (('v5'+k, k+'areru') for k in v5c))

    e('causative', (('v1', 'saseru'), ('v5u', 'waseru')),
      (('v5'+k, k+'aseru') for k in v5c))

    e('prohibitive', (('v1', 'runa'), ('v5u', 'una')),
      (('v5'+k, k+'una') for k in v5c))

    e('imperative', (('v1', 'ro'), ('v5u', 'e')),
      (('v5'+k, k+'e') for k in v5c))

    return {k:to_hiragana(v) for k,v in d.items()}

conjugation_table = _make_conjugation_table()

def _print_conjugation_table():
    from tabulate import tabulate
    d = conjugation_table
    forms = tuple(sorted(set(form for form,stem in d.keys())))
    stems = tuple('v5'+a for a in 'utrkgbms') + ('v1',)
    F = to_kunrei
    return tabulate(tuple((s,)+tuple(F(d[(f, s)]) for f in forms)
                          for s in stems),
                    headers=("stem",)+tuple(f[:3] for f in forms),
                    tablefmt='simple')

def _make_potential_conjugations_table(forms_condition):
    d = {}
    econ = ('',) + tuple(con)
    with_u = tuple(('る',))
    with_u_re = re.compile("{}$".format(re_one_of(with_u)))
    for stems, s, prefixlen in chain(
            ((('v5u',)     ,      'u', 0),),
            ((('v5'+k,)    ,   k+ 'u', 0) for k in 'trkgbms'),
            ((('v1', 'v5r'), c+v+'ru', 1) for c in econ for v in 'ie')):
        a = frozenset(v for (form, stem), v
                      in conjugation_table.items()
                      if stem in stems and forms_condition(form))
        d[to_hiragana(s)] = (a | frozenset(with_u_re.sub(x, '') for x in a),
                             prefixlen)
    return d

potential_conjugations_table = _make_potential_conjugations_table(lambda form: True)

potential_pol_conjugations_table = _make_potential_conjugations_table(
    lambda form: form=='polite')

recognize_conjugation_re = re.compile("({})$".format(
    re_one_of(potential_conjugations_table.keys())))
def recognize_conjugation(reading, okurigana, postreading):
    ''' returns n where n is the number of matching
characters, or None if nonmatching '''
    if not okurigana:
        return None
    full = reading+okurigana
    m = recognize_conjugation_re.search(full)
    if not m:
        return None
    c = potential_conjugations_table.get(m.group(1), None)
    if c is None:
        return None
    conj, prefixlen = c
    okurigana_stem_end = (m.start(1) + prefixlen) - len(reading)
    if okurigana_stem_end < 0:
        return None
    postreading_conj = postreading[okurigana_stem_end:]
    for i in range(min(len(postreading_conj), 5), -1, -1):
        s = postreading_conj[:i]
        if s in conj:
            return i
    return None

# from wikipedia
rendaku_map = {'k': ('g',),
               's': ('z','j'),
               't': ('z','j','d'),
               'h': ('b','p')} # added 'p'; correct?

rendaku_actual_map = {to_hiragana(c+v):tuple(set(to_hiragana(nc+v)
                                                 for nc in ncs))
                      for c, ncs in rendaku_map.items()
                      for v in vow}

assert all(len(k)==1 for k in rendaku_actual_map)

def rendaku_possibilities(word):
    '''note: this does not return the original word'''
    c = word[0]
    replacements = rendaku_actual_map.get(c, None)
    if replacements is None:
        return ()
    else:
        suf = word[1:]
        return (r+suf for r in replacements)

split_mora_longest_re = re.compile('{}'.format(re_one_of(all_mora.values())))
def split_mora_longest(s):
    prev_end = 0
    r = []
    for m in re.finditer(split_mora_longest_re, s):
        if prev_end is not None:
            if m.start(1) != prev_end:
                break
        r.append(m.group(1))
        prev_end = m.end(1)
    return (prev_end, r)

def make_similarity_table(metric):
    d = {(xv, yv): metric(xk, yk)
         for xk, xv in all_mora.items()
         for yk, yv in all_mora.items()}

default_similarity_metric_strict = set(("n'", 'xtu'))
def default_similarity_metric(x, y):
    if x == y:
        return 2
    if set((x, y)) & default_similarity_metric_strict:
        return 0
    if x[0] == y[0]:
        return 1
    return 0

default_similarity_table = make_similarity_table(default_similarity_metric)
def similarity(a, b):
    '''returns (index_in_a, index_in_b)'''
    return (ae, be, (default_similarity_table.get(x_y, 0)
                     for x_y in zip(ar, br)))



from .mora import (rendaku_possibilities,
                   recognize_conjugation)
import romkan
from pyjmdict.db.kanjidic import Character
from collections import namedtuple, OrderedDict
from itertools import chain
import functools
import numpy as np

try:
    import xml.etree.ElementTree as ElementTree
except ImportError:
    ElementTree = None

ja_readings = frozenset(('ja_on', 'ja_kun'))
def all_readings(character):
    # TODO: SQL
    for rmg in character.rms:
        yield from (x for x in rmg.readings
                    if x.type in ja_readings)

class BaseFragment(object):
    def get_costs(self, word, preword):
        '''must yield (partial_cost, length, result)
if word is "abcdef" and we are evaluating cost at i=2, then
this function will receive word="cdef" and preword="ab"'''
        return ()

def common_prefix_length(x, y):
    i = 0
    for i, (a, b) in enumerate(zip(x, y)):
        if a != b:
            break
    else:
        return i + 1
    return i

class Fragment(BaseFragment):
    readings = ()
    skip_costs = ()
    source_str = ...
    kanji_postword = ...
    kanji_preword  = ...
    def __init__(self, readings, skip_costs, source_str):
        self.readings   = readings
        self.skip_costs = skip_costs
        self.source_str = source_str
    def get_costs(self, word, preword):
        for r in self.readings:
            reading, cost = r.reading, r.cost
            if word.startswith(reading):
                if reading:
                    postreading = self.kanji_postword
                    z = common_prefix_length(r.okurigana, postreading)
                    n = recognize_conjugation(reading, r.okurigana,
                                              postreading)
                    m = recognize_conjugation(reading, r.okurigana,
                                              word[len(reading):])
                    if (not r.okurigana) or ((n or m) is not None) or z:
                        cost_ = (cost - z*2000 - (n or m or 0)*1000
                                 - 10*((n or m) is not None))
                        yield (len(reading), cost_, r)
                        if m:
                            yield (len(reading)+m, cost_,
                                   r._replace(reading=word[:len(reading)+m]))
                else:
                    yield (len(reading), cost, r)
            if (word.startswith('の') and preword.endswith('ん')
                and reading.startswith('お')): # 天皇
                reading2 = reading.replace('お', 'の', 1)
                if word.startswith(reading2):
                    yield (len(reading), cost,
                           r._replace(reading=reading2))
        for skip_length, cost in self.skip_costs:
            if skip_length <= len(word):
                yield (skip_length, cost, word[:skip_length])
        yield from super().get_costs(word, preword)

def explain_reading_algorithm(fragments, word):
    '''FIXME TODO documentation
'''
    len_fragments = len(fragments)
    len_word = len(word)

    # memoization is lazy man's dynamic programming
    @functools.lru_cache(maxsize=None)
    def f(frag_index, word_index):
        if frag_index == len_fragments and word_index == len_word:
            return (0, ())
        if frag_index >= len_fragments or  word_index >  len_word:
            return (np.inf, ())
        args = (word[word_index:],
                word[:word_index])
        results = []
        for length, cost, result in fragments[frag_index].get_costs(*args):
            tail_cost, tail_results = f(frag_index + 1,
                                        word_index + length)
            results.append((tail_cost + cost, (result,) + tail_results))
        if not len(results):
            return (np.inf, ())
        best = min(results, key=lambda x:x[0])
        return best
    return f(0, 0)

def link_fragments(fragments):
    '''note: modifies fragments in-place'''
    ss = tuple(normalize_to_hiragana(frag.source_str)
               for frag in fragments)
    acc = ''
    for frag, s in zip(fragments, ss):
        frag.kanji_preword = acc
        acc = acc + s
    acc = ''
    for frag, s in zip(reversed(fragments), reversed(ss)):
        frag.kanji_postword = acc
        acc = acc + s
    return fragments

identity = lambda x:x

def uniq(iterable, key=identity):
    seen = set()
    for x in iterable:
        k = key(x)
        if k not in seen:
            seen.add(k)
            yield x

@functools.lru_cache(maxsize=256)
def normalize_to_hiragana(s):
    return romkan.to_hiragana(romkan.to_kunrei(s))

ReadingPossibility = namedtuple(
    "ReadingPossibility",
    ['cost', 'reading', 'okurigana',
     'source', 'original'])

def split_okurigana(s):
    spl = s.split('.', 1)
    if len(spl) > 1:
        return tuple(spl)
    else:
        return (s, '')

def expand_possibilities(source, base_cost):
    def stage0():
        s = str(source)
        s = s.strip('-')
        reading, okurigana = split_okurigana(s)
        if okurigana:
            yield ReadingPossibility(
                base_cost, reading, okurigana,
                source, s)
        else:
            yield ReadingPossibility(
                base_cost, reading, '',
                source, s)

    def stage1():
        for r in stage0():
            r = r._replace(
                reading=normalize_to_hiragana(r.reading),
                okurigana=normalize_to_hiragana(r.okurigana))
            yield r
            for r_rendaku in rendaku_possibilities(r.reading):
                yield r._replace(reading=r_rendaku, cost=r.cost+50)

    def stage2():
        for r in stage1():
            yield r
            s = r.reading
            if s.endswith('つ') or s.endswith('く'):
                yield r._replace(reading=s[:-1]+'っ', cost=r.cost+100)

    yield from stage2()

ReadingAnnotation = namedtuple(
    "ReadingAnnotation",
    ['kanji', 'reading', 'okurigana', 'source', 'original'])

def explain_reading(chars, target_reading, override):
    if not len(chars): return ()
    penalty = tuple((i, (i+1)*(1000000-i)) for i in range(0,11))

    readings = (
        (((ReadingPossibility(cost=0, reading=c, okurigana='',
                              source=None, original=c),)
          if isinstance(c, str) else
          chain.from_iterable(
              (expand_possibilities(r, i*5))
              for i, r in enumerate(all_readings(c))))
         if j not in override else
         override[j])
        for j,c in enumerate(chars))

    # reading_dicts = tuple(reading_dicts)
    # print(reading_dicts[1], reading_dict_override)
    # print([list(x.keys()) for x in reading_dicts])

    readings = tuple(tuple(
        uniq(sorted(rs, key=lambda r:r.cost),
             key=lambda r:(r.reading, r.okurigana)))
                     for rs in readings)

    score, split = explain_reading_algorithm(link_fragments(tuple(
        Fragment(readings=rs,
                 skip_costs=(() if isinstance(c, str) else penalty),
                 source_str=str(c))
        for c,rs in zip(chars, readings))), target_reading)

    def process_single_result(char, r):
        if isinstance(r, str): # from skip entry
            return ReadingAnnotation(
                kanji     = str(char),
                reading   = r,
                okurigana = '',
                source    = None,
                original  = None)
        else:
            return ReadingAnnotation(
                kanji     = str(char),
                reading   = r.reading,
                okurigana = r.okurigana,
                source    = r.source,
                original  = r.original)

    return tuple(process_single_result(c, r)
                 for c, r in zip(chars, split))

def joinlist(inter, iterable):
    iterable = iter(iterable)
    yield next(iterable)
    for x in iterable:
        yield from inter
        yield x

def explain_reading_helper(S, word, reading):
    '''\
returns ReadingAnnotation'''
    q = S.query(Character)
    l = tuple((q.get(c) or c, None) for c in word)
    inter = ('', tuple(ReadingPossibility(cost=cost,
                                          reading=r,
                                          okurigana='',
                                          source=None,
                                          original='')
                       for r, cost in (('', 0),
                                       ('の', 10000))))
    l = tuple(joinlist((inter,), l))
    return explain_reading(
        tuple(x[0] for x in l), reading,
        {i: override for i, (c, override) in enumerate(l)
         if override is not None})

class _NoSourceReading(object):
    def __repr__(self):
        return '<NoSourceReading>'
    def __str__(self):
        return '×'

no_source_reading = _NoSourceReading()

ReadingColumn = namedtuple(
    'ReadingColumn',
    ('kanji', 'reading', 'source_reading', 'source_okurigana', 'source_red'))

def count_True_after(lst):
    r = []
    n = len(lst)
    j = n
    for i in reversed(range(n)):
        r.insert(0, j-i-1)
        if not lst[i]:
            j = i
    r.insert(0, j)
    return tuple(r)

def explain_reading_to_reading_columns(result):
    def process(R):
        c,r,src,orig = R.kanji, R.reading, R.source, R.original
        if orig is not None:
            o_r, o_oku = split_okurigana(normalize_to_hiragana(orig))
        else:
            o_r, o_oku = '', ''
        if not c:
            f = ReadingColumn('', r, '', '', False)
        elif orig is None:
            f = ReadingColumn(c, r, no_source_reading, '', True)
        elif orig == c:
            f = ReadingColumn(c, '', '', '', False)
        elif o_r == r:
            if o_oku:
                f = ReadingColumn(c, r, o_r, o_oku, False)
            else:
                f = ReadingColumn(c, r, '', '', False)
        else:
            f = ReadingColumn(c, r, o_r, o_oku, True)
        return f
    return tuple(process(R) for R in result if R.kanji or R.reading)

def explain_reading_to_text(result):
    cols = explain_reading_to_reading_columns(result)

    r = []
    a = r.append
    for R in cols:
        a(R.kanji)
        if R.reading:
            a("[")
            if R.source_reading or R.source_okurigana:
                o_r, oku, red = (str(R.source_reading), R.source_okurigana,
                                 R.source_red)
                if R.source_red: a("!")
                a(o_r)
                if oku: a('⋅'+oku)
                a("→")
            a(R.reading)
            a("]")
    return ''.join(r)

def explain_reading_to_html(result, class_prefix='fgn-',
                            ElementTree=ElementTree):
    def td(text, attrib=()):
        e = ElementTree.Element('td')
        e.text = text
        e.attrib.update(attrib)
        return e

    cssp = class_prefix

    tr = [ElementTree.Element('tr') for k in range(3)]
    cols = explain_reading_to_reading_columns(result)
    tr2_empty = [not (R.source_reading or R.source_okurigana) for R in cols]
    tr2_empty_after = count_True_after(tr2_empty)
    pre = tr2_empty_after[0]
    if pre > 0:
        tr[2].append(td('', {'colspan': str(pre)}))
    for i, R in enumerate(cols):
        color_class = cssp+'c{}'.format(i%3)
        k = R.kanji
        tr[0].append(td(k if k else '↓',
                        {'class': color_class if k else cssp+'extra'}))
        tr[1].append(td(R.reading, {'class': color_class}))
        if not tr2_empty[i]:
            o_r, oku, red = (str(R.source_reading), R.source_okurigana,
                             R.source_red)
            source_color_class = cssp+'red' if red else color_class
            append1, append2 = None, None
            colspan = tr2_empty_after[i+1]
            under = o_r
            if colspan > 0:
                d = {}
                if colspan > 1:
                    d['colspan'] = str(colspan)
                if oku:
                    d['class'] = cssp+'oku '+source_color_class
                append2 = (td('⋅'+oku if oku else None, d))
            elif oku: # no room for okurigana, ugh
                under = o_r+'⋅'+oku
            append1 = td(under, {'class': source_color_class})
            for a in (append1, append2):
                if a is not None:
                    tr[2].append(a)
    table = ElementTree.Element('table')
    table.extend(tr)
    table.attrib['class'] = cssp+'table'
    return (table,)

def explain_reading_to_html_css(class_prefix='fgn-'):
    return '''\
table.{p}table {{
vertical-align: top;
display: inline;
border-spacing: 0px;
border: none;
}}

table.{p}table tr {{
padding:0px;
margin:0px;
}}

table.{p}table td {{
padding:0px;
margin:0px;
text-align:center;
}}

table.{p}table tr:nth-child(1), table.{p}table tr:nth-child(1) > a {{
font-size:160%;
font-weight:normal;
color:black;
}}

table.{p}table td.{p}c0, table.{p}table td.{p}c0 > a {{
color:black;
}}

table.{p}table td.{p}c1, table.{p}table td.{p}c1 > a {{
color:green;
}}

table.{p}table td.{p}c2, table.{p}table td.{p}c2 > a {{
color:blue;
}}

table.{p}table tr:nth-child(2), table.{p}table tr:nth-child(3) {{
font-size: 100%;
}}

table.{p}table td.{p}extra, table.{p}table td.{p}extra > a {{
color:red;
font-size: 75%;
vertical-align: bottom;
}}

table.{p}table td.{p}red, table.{p}table td.{p}red > a {{
color:darkred;
}}

table.{p}table td.{p}oku {{
text-align: left;
}}
'''.format(p=class_prefix)



from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.associationproxy import association_proxy, _AssociationList
from sqlalchemy import (String, Unicode, Integer, Boolean, Index,
                        ForeignKey, ForeignKeyConstraint, UniqueConstraint)
from sqlalchemy import func, and_, or_
from sqlalchemy.orm import exc, relationship, aliased
import os
from jikan_sqlalchemy_utils.session import SessionMaker
from jikan_sqlalchemy_utils.combine_args import make_combine_args_class
from jikan_sqlalchemy_utils.ocolumn import OColumn
from jikan_sqlalchemy_utils.table import repr_helper, index
from jikan_sqlalchemy_utils.query import rel_to_query, pk_join_expr
from contextlib import closing
import functools

import tqdm

from pyjmdict.db.jmdict   import KanjiEntry, ReadingEntry
from pyjmdict.db.kanjidic import Character

from .explain_reading_db import ExplainReadingCache
from .explain_reading import (explain_reading_to_text,
                              normalize_to_hiragana)

import sqlalchemy as sa
from concurrent.futures import ProcessPoolExecutor, as_completed


Base = declarative_base()
CA = make_combine_args_class(Base)

class WordExample(CA, Base):
    __tablename__ = 'kwe_word_example'

    id = OColumn(Integer, primary_key=True, _order=0)
    kanji   = OColumn(Unicode(  2), _order=10)
    word    = OColumn(Unicode( 64), _order=11)
    reading = OColumn(Unicode(128), _order=12)

    @declared_attr
    def __partial_table_args__(cls):
        return (index(cls, 'kanji'),
                index(cls, 'word'),
                index(cls, 'reading'))

def get_examples(pyjmdict_session,
                 explain_reading_cache, kanji):
    S = pyjmdict_session

    # c = S.query(Character).get(kanji)

    k1 = aliased(KanjiEntry, name='k1')
    sq = S.query(k1, func.count(ReadingEntry.entry_id).label('readings_count'))

    sq = (sq
          # .filter(k1.pri_nf != None)
          .join(k1.readings).group_by(k1)
          # .with_parent(c, 'kanji_entries') # doesn't work
          .filter(k1.str.like('%'+kanji+'%')) # FIXME: FUCKING STUPID!
          .subquery())

    k2 = aliased(KanjiEntry, name='k2')
    q3 = (S.query(k2, sq.c.readings_count)
          .select_from(sq)
          .join(k2, pk_join_expr(k2, sq))
          .filter(sq.c.readings_count == 1)
          .order_by(k2.pri_nf)
          .limit(200))

    seen_readings = set()
    good = []
    for ke, _ in q3.all():
        word, word_reading = str(ke), str(ke.readings[0])

        if S.query(KanjiEntry).filter(KanjiEntry.str == str(ke)).count() > 1:
            continue

        result = explain_reading_cache(pyjmdict_session,
                                       word, word_reading)

        # reject any irregular readings
        if any((a.kanji or a.reading) and a.source is None
               for a in result):
            continue

        try:
            ann = next(a for a in result if a.kanji==kanji)
        except StopIteration:
            continue

        reading = normalize_to_hiragana(ann.source.strip('-').partition('.')[0])
        if reading not in seen_readings:
            # print(reading, word, seen_readings, ann)
            seen_readings.add(reading)
            good.append(WordExample(
                kanji=kanji, word=word, reading=word_reading))

    return good

@functools.lru_cache(maxsize=8)
def init_pyjmdict(fn):
    return SessionMaker.from_uri_filename(
        fn, isolation_level='READ UNCOMMITTED')

@functools.lru_cache(maxsize=8)
def init_erc(fn):
    csm = SessionMaker.from_uri_filename(
        fn, isolation_level='READ COMMITTED')

    return ExplainReadingCache(cache_sessionmaker=csm)

@functools.lru_cache(maxsize=8)
def init_out(fn):
    osm = SessionMaker.from_uri_filename(
        fn, isolation_level='SERIALIZABLE')

    return osm

def do_db_init(args):
    osm = init_out(args.kwe_db)
    osm.populate(Base)

def do_kanji(args, kanji):
    psm = init_pyjmdict(args.pyjmdict_db)
    osm = init_out(args.kwe_db)
    erc = init_erc(args.erc_db)

    with closing(psm()) as sess:
        if sess.query(Character).get(kanji) is None:
            return

    for i in range(10):
        try:
            with closing(osm()) as out_session:
                if out_session.query(WordExample).filter(
                        WordExample.kanji == kanji).count():
                    return
            with closing(psm()) as pyjmdict_session:
                exs = list(get_examples(
                    pyjmdict_session=pyjmdict_session,
                    explain_reading_cache=erc,
                    kanji=kanji))
            with closing(osm()) as out_session:
                for ex in exs:
                    out_session.add(ex)
                out_session.commit()
        except sa.exc.ProgrammingError:
            pass
        else:
            break

def main():
    import argparse
    parser = argparse.ArgumentParser(
        description=("generate kanji word example db"))
    parser.add_argument('--kanjis', required=True,
                        help='file containing kanjis')
    parser.add_argument('-j', '--jobs', type=int, default=None,
                        help='Allow N jobs at once.')
    parser.add_argument('-u', '--pyjmdict-db', required=True,
                        help='file from which to read pyjmdict database URI')
    parser.add_argument('-e', '--erc-db', required=True,
                        help='file from which to read explain-'
                        'reading cache database URI')
    parser.add_argument('-o', '--kwe-db', required=True,
                        help='file from which to read output database URI')
    args = parser.parse_args()

    kanjis = set()
    with open(args.kanjis, 'rt') as handle:
        while True:
            block = handle.read(1000)
            if not block: break
            kanjis.update(block)
            # break

    kanjis = list(sorted(kanjis))

    executor = ProcessPoolExecutor(max_workers=args.jobs)

    # init db
    executor.submit(do_db_init, args).result()

    futures = {executor.submit(do_kanji, args, kanji):
               kanji for kanji in kanjis}
    for future in tqdm.tqdm(as_completed(futures), total=len(futures)):
        kanji = futures[future]
        try:
            future.result()
        except Exception as exc:
            print('{!r} generated an exception'.format(kanji))
            raise

if __name__ == '__main__':
    main()





class BaseUnicodeRanges(object):
    def regex_string(self):
        return '[{}]'.format(''.join(
            '\\u{:04x}-\\u{:04x}'.format(start, end)
            for start, end in self.ranges))

class CJKUnicodeRanges(BaseUnicodeRanges):
    # TODO: add comments for each section source
    ranges = ((0x2e80, 0x2fd5),
              (0x2ff0, 0x2ffb),
              (0x3000, 0x3020),
              (0x3040, 0x309f),
              (0x30a0, 0x30ff),
              (0x31f0, 0x31ff),
              (0x3400, 0x4dbf),
              (0x4e00, 0x9faf),
              (0xff01, 0xff20),
              (0xff3b, 0xff40),
              (0xff5b, 0xff65))




import xml.etree.ElementTree as ET
from .explain_reading import explain_reading_to_html, explain_reading_helper
import etreetools.lib as EL

def html_make_list_if_more_than_one(l, tagname='ol'):
    n = len(l)
    if n <= 1:
        return l
    else:
        ol = ET.Element(tagname)
        for x in l:
            li = ET.Element('li')
            EL.extend(li, x)
            ol.append(li)
        return (ol,)

def html_word_entry(S, kanji_entry):
    ke = kanji_entry
    k = str(ke)
    reading = str(ke.readings[0])
    senses = ke.senses

    eg = html_make_list_if_more_than_one(tuple(
        filter(len,
               ('; '.join(str(gloss) for gloss in se.glosses
                         if gloss.lang == 'eng')
                for se in senses))))
    return eg

def html_make_p(*args):
    p = ET.Element('p')
    for a in args:
        EL.extend(p, a)
    return p



from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.associationproxy import association_proxy, _AssociationList
from sqlalchemy import (String, Unicode, Integer, Boolean, Index,
                        ForeignKey, ForeignKeyConstraint, UniqueConstraint)
from sqlalchemy import func, and_, or_
from sqlalchemy.orm import exc, relationship
import os
from jikan_sqlalchemy_utils.combine_args import make_combine_args_class
from jikan_sqlalchemy_utils.ocolumn import OColumn
from jikan_sqlalchemy_utils.table import repr_helper, index

from .explain_reading import ReadingAnnotation, explain_reading_helper

Base = declarative_base()
CA = make_combine_args_class(Base)

class Word(CA, Base):
    __tablename__ = 'er_word'

    id = OColumn(Integer, primary_key=True, _order=0)
    str     = OColumn(Unicode( 32), _order=10)
    reading = OColumn(Unicode(128), _order=11)

    annotations = relationship('Annotation', order_by='Annotation.i',
                               back_populates='word')

    @declared_attr
    def __partial_table_args__(cls):
        return (index(cls, 'str'),
                index(cls, 'reading'))

    def __repr__(self):
        return repr_helper('W', (
            (None, str(self)),
            ('reading', self.readings)))

class Annotation(CA, Base):
    __tablename__ = 'er_annotation'

    word_id = OColumn(Integer, ForeignKey(Word.__tablename__+".id"),
                      primary_key=True, _order=0)
    i = OColumn(Integer,
                primary_key=True, nullable=False, _order=1)

    kanji   = OColumn(Unicode(16), _order=10)
    reading = OColumn(Unicode(32), _order=11)
    okurigana = OColumn(Unicode(32), _order=12)
    source    = OColumn(Unicode(32), _order=13)
    original  = OColumn(Unicode(32), _order=14)

    word = relationship('Word', back_populates='annotations')

    @declared_attr
    def __partial_table_args__(cls):
        return (index(cls, 'kanji'),
                index(cls, 'reading'),
                index(cls, 'okurigana'),
                index(cls, 'source'),
                index(cls, 'original'))

    def __repr__(self):
        return repr_helper('R', (
            (None, self.kanji),
            (None, self.reading),
            ('okurigana', self.okurigana),
            ('source', self.source),
            ('original', self.original)))

    def to_reading_annotation(self):
        return ReadingAnnotation(
            **{k: getattr(k) for k in
               ReadingAnnotation._fields})

class ExplainReadingCache:
    def __init__(self, cache_sessionmaker):
        self.cache_sessionmaker = cache_sessionmaker
        self.populate_db()

    def populate_db(self):
        Base.metadata.create_all(self.cache_sessionmaker().bind)

    def actual_call(self, session, word, reading):
        return explain_reading_helper(session, word, reading)

    def __call__(self, session, word, reading):
        ses = self.cache_sessionmaker()
        try:
            w = ses.query(Word).filter(
                and_(Word.str == word,
                     Word.reading == reading)).first()

            if w is None:
                result = self.actual_call(session, word, reading)
                w = Word(str=word, reading=reading)
                ses.add(w)

                for a in (Annotation(
                    word=w, i=i,
                    kanji=ann.kanji,
                    reading=ann.reading,
                    okurigana=ann.okurigana,
                    source=None if ann.source is None else str(ann.source),
                    original=ann.original)
                              for i, ann in enumerate(result)):
                    ses.add(a)

            ses.commit()
            return list(w.annotations)
        finally:
            ses.close()




from jikan_sqlalchemy_utils.session import SessionMaker
from pyjmdict.db.jmdict import *
from pyjmdict.db.kanjidic import *
import sqlalchemy as sa
from sqlalchemy import func, and_, or_
from sqlalchemy.orm import aliased
from sqlalchemy.inspection import inspect
from .sau_fixed import get_primary_keys
import json
from collections import OrderedDict, namedtuple
from .explain_reading import (explain_reading_helper, explain_reading_to_text,
                              explain_reading_to_html,
                              explain_reading_to_html_css)
from .explain_reading_db import ExplainReadingCache
import xml.etree.ElementTree as ET
from .jmdict import html_make_p, html_word_entry

def main():
    with open('Heisigs_RTK_6th.json') as handle:
        heisig = json.load(handle)
    sm = SessionMaker.from_uri_filename('database.txt')

    csm = SessionMaker.from_sqlite_filename(
        "cache.sqlite", isolation_level='SERIALIZABLE')

    erc = ExplainReadingCache(cache_sessionmaker=csm)

    S = sm()
    qke = S.query(KanjiEntry)
    q = S.query(KanjiEntry).filter(KanjiEntry.pri_nf <= 40)
    q = q.order_by(func.length(KanjiEntry.str).desc())
    q = q.limit(1200)
    print(HEADER)

    for ke in q.all():
        print(ET.tostring(html_make_p(
            explain_reading_to_html(
                erc(S, str(ke), str(ke.readings[0])))
            +(' ',)+html_word_entry(S, ke)), 'unicode'))
    print('''</body>
</html>
''')

HEADER = '''<html lang="en">
  <meta charset="utf-8">
  <head>
    <style>{}

body {{
font-family: "Noto Sans", "Noto Sans CJK JP", sans-serif;
}}
</style>
  </head>
  <body>
'''.format(explain_reading_to_html_css())

if __name__=='__main__':
    main()




import re
from itertools import chain

def re_one_of(xs):
    """return a regex string that matches one of the xs"""
    # the "a^" thing matches nothing on purpose (not even with re.M).
    # it is necessary in the case where `xs` is empty because then the
    # regex is "" which then matches the empty string -- oops.
    escape = re.escape
    xs = sorted(xs, key=len, reverse=True)
    return '|'.join(escape(x) for x in xs) if xs else 'a^'



from setuptools import setup, find_packages
pkg = "learnusumjap"
ver = '0.1.19'
setup(
    name             = pkg,
    version          = ver,
    description      = "Japanese language utilities",
    author           = "jikan@cock.li",
    author_email     = "jikan@cock.li",
    license          = "LGPLv3",
    url              = "https://hydra.ecd.space/jikan/learnusumjap/",
    packages         = find_packages(),
    install_requires = ['pyjmdict>=0.4.2',
                        'jikan_sqlalchemy_utils>=0.0.4',
                        'romkan>=0.2.1',
                        'tqdm'],
    classifiers      = ["Programming Language :: Python :: 3 :: Only"])
